# Copyright 2018 Rackspace, US Inc.
# Copyright 2020 A10 Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import sys

from oslo_db.sqlalchemy import enginefacade
import oslo_i18n as i18n
from oslo_log import log as logging

from a10_nlbaas2oct import db_utils


def migrate_vip_ports(n_session, oct_accnt_id, lb_id, n_lb):

    # Migrate the port and security groups to Octavia
    vip_port = n_session.execute(
        "SELECT a.device_owner, a.project_id, b.security_group_id "
        "FROM ports a JOIN securitygroupportbindings b ON "
        "a.id = b.port_id  where id = :id;",
        {'id': n_lb[7]}).fetchone()
    # neutron-lbaas does not support user VIP ports, so take
    # ownership of the port and security group
    if vip_port[0] == 'neutron:LOADBALANCERV2':
        result = n_session.execute(
            "UPDATE ports SET device_owner = 'Octavia', "
            "project_id = :proj_id WHERE "
            "id = :id;", {'id': n_lb[7],
                          'proj_id': oct_accnt_id})
        if result.rowcount != 1:
            raise Exception(_('Unable to update VIP port in the neutron '
                            'database.'))
        security_group = n_session.execute(
            "SELECT project_id FROM securitygroups WHERE id = :id",
            {'id': vip_port[2]}).fetchone()
        # Update security group project, only when its owner is not the
        # user project, which means that Octavia should own it
        if security_group[0] != n_lb[1]:
            result = n_session.execute(
                "UPDATE securitygroups SET project_id = :proj_id WHERE "
                "id = :id;", {'proj_id': oct_accnt_id,
                              'id': vip_port[2]})
            if result.rowcount != 1:
                raise Exception(_('Unable to update VIP security group in '
                                  'the neutron database.'))

def migrate_lb(o_session, lb_id, n_lb, fl_id):
    """ 
        Create the load balancer. Provider name is hardcoded.
        This tool is only meant for A10 LB migration. On nlbaas the provider is 'a10networks'
        but on octavia the provider is 'a10'
    """
    provider_name = "a10"

    result = o_session.execute(
        "INSERT INTO load_balancer (id, project_id, name, "
        "description, provisioning_status, operating_status, enabled, "
        "created_at, updated_at, flavor_id, provider) VALUES (:id, :project_id, "
        ":name, :description, :provisioning_status, "
        ":operating_status, :enabled, :created_at, :updated_at, "
        ":flavor_id, :provider);",
        {'id': lb_id, 'project_id': n_lb[1], 'name': n_lb[2],
         'description': n_lb[3], 'provisioning_status': 'ACTIVE',
         'operating_status': n_lb[5], 'enabled': n_lb[4],
         'created_at': datetime.datetime.utcnow(),
         'updated_at': datetime.datetime.utcnow(),
         'flavor_id': fl_id, 'provider': provider_name,})
    if result.rowcount != 1:
          raise Exception(_('Unable to create load balancer in the '
                          'Octavia database.'))


def migrate_vip(n_session, o_session, lb_id, n_lb):
    # Get the network ID for the VIP
    subnet = n_session.execute(
        "SELECT network_id FROM subnets WHERE id = :id;",
        {'id': n_lb[8]}).fetchone()
    # Create VIP record
    result = o_session.execute(
        "INSERT INTO vip (load_balancer_id, ip_address, port_id, "
        "subnet_id, network_id) VALUES (:lb_id, :ip_address, "
        ":port_id, :subnet_id, :network_id);",
        {'lb_id': lb_id, 'ip_address': n_lb[9], 'port_id': n_lb[7],
         'subnet_id': n_lb[8], 'network_id': subnet[0]})
    if result.rowcount != 1:
        raise Exception(_('Unable to create VIP in the Octavia '
                        'database.'))


def migrate_listener(n_session, o_session, lb_id, n_lb, listener, lb_stats):
    # Create listener
    result = o_session.execute(
        "INSERT INTO listener (id, project_id, name, description, "
        "protocol, protocol_port, connection_limit, "
        "load_balancer_id, tls_certificate_id, default_pool_id, "
        "provisioning_status, operating_status, enabled, "
        "created_at, updated_at) VALUES (:id, :project_id, :name, "
        ":description, :protocol, :protocol_port, "
        ":connection_limit, :load_balancer_id, "
        ":tls_certificate_id, :default_pool_id, "
        ":provisioning_status, :operating_status, :enabled, "
        ":created_at, :updated_at);",
        {'id': listener[0], 'project_id': n_lb[1],
         'name': listener[1], 'description': listener[2],
         'protocol': listener[3], 'protocol_port': listener[4],
         'connection_limit': listener[5],
         'load_balancer_id': lb_id,
         'tls_certificate_id': listener[10],
         'default_pool_id': listener[6],
         'provisioning_status': listener[8],
         'operating_status': listener[9], 'enabled': listener[7],
         'created_at': datetime.datetime.utcnow(),
         'updated_at': datetime.datetime.utcnow()})
    if result.rowcount != 1:
        raise Exception(_('Unable to create listener in the '
                        'Octavia database.'))
    # Convert load balancer stats to listener stats
    # This conversion may error on the low side due to
    # the division
    listeners, lb_stats = db_utils.get_listeners_and_stats_by_lb(n_session, lb_id)
    result = o_session.execute(
        "INSERT INTO listener_statistics (listener_id, bytes_in, "
        "bytes_out, active_connections, total_connections, "
        "amphora_id, request_errors) VALUES (:listener_id, "
        ":bytes_in, :bytes_out, :active_connections, "
        ":total_connections, :amphora_id, :request_errors);",
        {'listener_id': listener[0],
         'bytes_in': int(lb_stats[0] / len(listeners)),
         'bytes_out': int(lb_stats[1] / len(listeners)),
         'active_connections': int(lb_stats[2] / len(listeners)),
         'total_connections': int(lb_stats[3] / len(listeners)),
         'amphora_id': listener[0], 'request_errors': 0})
    if result.rowcount != 1:
        raise Exception(_('Unable to create listener statistics '
                        'in the Octavia database.'))


def migrate_SNI(o_session, listener_id, SNI):
    # Create SNI containers
    result = o_session.execute(
        "INSERT INTO sni (listener_id, tls_container_id, position) VALUES "
        "(:listener_id, :tls_container_id, :position);",
        {'listener_id': listener_id, 'tls_container_id': SNI[0],
            'position': SNI[1]})
    if result.rowcount != 1:
        raise Exception(_('Unable to create SNI record in the Octavia '
                        'database.'))


def migrate_l7policy(o_session, project_id, listener_id, l7policy):
        # Create L7 Policies
        L7p_op_status = 'ONLINE' if l7policy[9] else 'OFFLINE'
        result = o_session.execute(
            "INSERT INTO l7policy (id, name, description, listener_id, "
            "action, redirect_pool_id, redirect_url, position, enabled, "
            "provisioning_status, created_at, updated_at, project_id, "
            "operating_status) VALUES (:id, :name, :description, "
            ":listener_id, :action, :redirect_pool_id, :redirect_url, "
            ":position, :enabled, :provisioning_status, :created_at, "
            ":updated_at, :project_id, :operating_status);",
            {'id': l7policy[0], 'name': l7policy[1],
             'description': l7policy[2], 'listener_id': listener_id,
             'action': l7policy[4], 'redirect_pool_id': l7policy[5],
             'redirect_url': l7policy[6], 'position': l7policy[7],
             'enabled': l7policy[9], 'provisioning_status': l7policy[8],
             'created_at': datetime.datetime.utcnow(),
             'updated_at': datetime.datetime.utcnow(),
             'project_id': project_id, 'operating_status': L7p_op_status})
        if result.rowcount != 1:
            raise Exception(_('Unable to create L7 policy in the Octavia '
                            'database.'))


def migrate_l7rule(o_session, project_id, l7policy, l7rule):
    # Create L7rule
    L7r_op_status = 'ONLINE' if l7rule[7] else 'OFFLINE'
    result = o_session.execute(
        "INSERT INTO l7rule (id, l7policy_id, type, compare_type, "
        "`key`, value, invert, provisioning_status, created_at, "
        "updated_at, project_id, enabled, operating_status) VALUES "
        "(:id, :l7policy_id, :type, :compare_type, :key, :value, "
        ":invert, :provisioning_status, :created_at, :updated_at, "
        ":project_id, :enabled, :operating_status);",
        {'id': l7rule[0], 'l7policy_id': l7policy[0],
         'type': l7rule[1], 'compare_type': l7rule[2],
         'key': l7rule[4], 'value': l7rule[5], 'invert': l7rule[3],
         'provisioning_status': l7rule[6],
         'created_at': datetime.datetime.utcnow(),
         'updated_at': datetime.datetime.utcnow(),
         'project_id': project_id, 'enabled': l7rule[7],
         'operating_status': L7r_op_status})
    if result.rowcount != 1:
        raise Exception(_('Unable to create L7 policy in the Octavia '
                        'database.'))

def migrate_pools(o_session, lb_id, n_lb, pool):
    # Create pools
    result = o_session.execute(
        "INSERT INTO pool (id, project_id, name, description, "
        "protocol, lb_algorithm, operating_status, enabled, "
        "load_balancer_id, created_at, updated_at, "
        "provisioning_status) VALUES (:id, :project_id, :name, "
        ":description, :protocol, :lb_algorithm, "
        ":operating_status, :enabled, :load_balancer_id,"
        ":created_at, :updated_at, :provisioning_status);",
        {'id': pool[0], 'project_id': n_lb[1], 'name': pool[1],
         'description': pool[2], 'protocol': pool[3],
         'lb_algorithm': pool[4], 'operating_status': pool[8],
         'enabled': pool[6], 'load_balancer_id': lb_id,
         'created_at': datetime.datetime.utcnow(),
         'updated_at': datetime.datetime.utcnow(),
         'provisioning_status': pool[7]})
    if result.rowcount != 1:
        raise Exception(_('Unable to create pool in the '
                        'Octavia database.'))


def migrate_health_monitor(o_session, project_id, pool_id, hm_id, hm):
    # Create health monitor

    hm_op_status = 'ONLINE' if hm[7] else 'OFFLINE'

    result = o_session.execute(
        "INSERT INTO health_monitor (id, project_id, pool_id, type, delay, "
        "timeout, fall_threshold, rise_threshold, http_method, url_path, "
        "expected_codes, enabled, provisioning_status, name, created_at, "
        "updated_at, operating_status) VALUES (:id, :project_id, :pool_id, "
        ":type, :delay, :timeout, :fall_threshold, :rise_threshold, "
        ":http_method, :url_path, :expected_codes, :enabled, "
        ":provisioning_status, :name, :created_at, :updated_at, "
        ":operating_status);",
        {'id': hm_id, 'project_id': project_id, 'pool_id': pool_id,
         'type': hm[0], 'delay': hm[1], 'timeout': hm[2],
         'fall_threshold': hm[10], 'rise_threshold': hm[3],
         'http_method': hm[4], 'url_path': hm[5], 'expected_codes': hm[6],
         'enabled': hm[7], 'provisioning_status': hm[8], 'name': hm[9],
         'operating_status': hm_op_status,
         'created_at': datetime.datetime.utcnow(),
         'updated_at': datetime.datetime.utcnow()})
    if result.rowcount != 1:
        raise Exception(_('Unable to create health monitor in the Octavia '
                        'database.'))


def migrate_session_persistence(o_session, pool_id, sp):
    # Setup session persistence if it is configured
    result = o_session.execute(
        "INSERT INTO session_persistence (pool_id, type, cookie_name) "
        "VALUES (:pool_id, :type, :cookie_name);",
        {'pool_id': pool_id, 'type': sp[0], 'cookie_name': sp[1]})
    if result.rowcount != 1:
        raise Exception(_('Unable to create session persistence in the '
                        'Octavia database.'))


def migrate_member(o_session, project_id, pool_id, member):
    # Create member
    result = o_session.execute(
        "INSERT INTO member (id, pool_id, project_id, subnet_id, "
        "ip_address, protocol_port, weight, operating_status, enabled, "
        "created_at, updated_at, provisioning_status, name, backup) "
        "VALUES (:id, :pool_id, :project_id, :subnet_id, :ip_address, "
        ":protocol_port, :weight, :operating_status, :enabled, "
        ":created_at, :updated_at, :provisioning_status, :name, :backup);",
        {'id': member[0], 'pool_id': pool_id, 'project_id': project_id,
         'subnet_id': member[1], 'ip_address': member[2],
         'protocol_port': member[3], 'weight': member[4],
         'operating_status': member[7], 'enabled': member[5],
         'created_at': datetime.datetime.utcnow(),
         'updated_at': datetime.datetime.utcnow(),
         'provisioning_status': member[6], 'name': member[8],
         'backup': False})
    if result.rowcount != 1:
        raise Exception(
            _('Unable to create member in the Octavia database.'))



# Copyright 2020 A10 Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_db.sqlalchemy import enginefacade
import oslo_i18n as i18n
from oslo_log import log as logging
from oslo_utils import uuidutils

import acos_client
from a10_nlbaas2oct import a10_config as a10_cfg
from a10_nlbaas2oct import a10_migration as aten2oct
from a10_nlbaas2oct import db_utils
from a10_nlbaas2oct import flavor_migration as nexpr2fl
from a10_nlbaas2oct import lbaas_migration as lb2oct

_translators = i18n.TranslatorFactory(domain='a10_nlbaas2oct')

# The primary translation function using the well-known name "_"
_ = _translators.primary

CONF = cfg.CONF

cli_opts = [
    cfg.BoolOpt('all', default=False,
                help='Migrate all load balancers'),
    cfg.StrOpt('lb-id',
               help='Load balancer ID to migrate'),
    cfg.StrOpt('project-id',
               help='Migrate all load balancers owned by this project'),
]

migration_opts = [
    cfg.BoolOpt('delete_after_migration', default=False,
                help='Delete the load balancer records from neutron-lbaas'
                     ' after migration'),
    cfg.BoolOpt('trial_run', default=False,
                help='Run without making changes.'),
    cfg.StrOpt('octavia_account_id', required=True,
               help='The keystone account ID Octavia is running under.'),
    cfg.StrOpt('neutron_db_connection',
               required=True,
               help='The neutron database connection string'),
    cfg.StrOpt('octavia_db_connection',
               required=True,
               help='The octavia database connection string'),
    cfg.StrOpt('a10_nlbaas_db_connection',
               required=False,
               help='The a10 nlbaas database connection string'),
    cfg.StrOpt('a10_oct_db_connection',
               required=False,
               help='The a10 octavia database connection string'),
    cfg.StrOpt('a10_config_path',
               required=True,
               help='Path to config.py file used by the A10 networks lbaas driver'),
]

cfg.CONF.register_cli_opts(cli_opts)
cfg.CONF.register_opts(migration_opts, group='migration')


def main():
    if len(sys.argv) == 1:
        print('Error: Config file must be specified.')
        print('a10_nlbaas2oct --config-file <filename>')
        return 1
    logging.register_options(cfg.CONF)
    cfg.CONF(args=sys.argv[1:],
             project='a10_nlbaas2oct',
             version='a10_nlbaas2oct 1.0')
    logging.set_defaults()
    logging.setup(cfg.CONF, 'a10_nlbaas2oct')
    LOG = logging.getLogger('a10_nlbaas2oct')
    CONF.log_opt_values(LOG, logging.DEBUG)

    if not CONF.all and not CONF.lb_id and not CONF.project_id:
        print('Error: One of --all, --lb_id, --project_id must be specified.')
        return 1

    if ((CONF.all and (CONF.lb_id or CONF.project_id)) or
            (CONF.lb_id and CONF.project_id)):
        print('Error: Only one of --all, --lb_id, --project_id allowed.')
        return 1

    neutron_context_manager = enginefacade.transaction_context()
    neutron_context_manager.configure(
        connection=CONF.migration.neutron_db_connection)
    n_session_maker = neutron_context_manager.writer.get_sessionmaker()
    n_session = n_session_maker(autocommit=False)

    octavia_context_manager = enginefacade.transaction_context()
    octavia_context_manager.configure(
        connection=CONF.migration.octavia_db_connection)
    o_session_maker = octavia_context_manager.writer.get_sessionmaker()
    o_session = o_session_maker(autocommit=False)

    if CONF.migration.a10_nlbaas_db_connection:
        a10_nlbaas_context_manager = enginefacade.transaction_context()
        a10_nlbaas_context_manager.configure(
            connection=CONF.migration.a10_nlbaas_db_connection)
        a10_nlbaas_session_maker = a10_nlbaas_context_manager.writer.get_sessionmaker()
        a10_nlbaas_session = a10_nlbaas_session_maker(autocommit=False)
    else:
        a10_nlbaas_session = n_session

    if CONF.migration.a10_oct_db_connection:
        a10_oct_context_manager = enginefacade.transaction_context()
        a10_oct_context_manager.configure(
            connection=CONF.migration.a10_oct_db_connection)
        a10_oct_session_maker = a10_oct_context_manager.writer.get_sessionmaker()
        a10_oct_session = a10_oct_session_maker(autocommit=False)
    else:
        a10_oct_session = o_session

    LOG.info('Starting migration.')

    a10_config = a10_cfg.A10Config(config_dir=CONF.migration.a10_config_path,
                                   provider="a10networks")

    # Translate the name expressions into an Octavia flavor
    LOG.info('Migrating name expressions to flavors')
    fl_id = None
    flavor_data = nexpr2fl.create_flavor_data(a10_config)
    if flavor_data:
        fp_id = nexpr2fl.create_flavorprofile(o_session, flavor_data)
        fl_id = nexpr2fl.create_flavor(o_session, fp_id)

    # Migrate the loadbalancers and their child objects
    failure_count = 0
    lb_id_list = db_utils.get_loadbalancer_ids(n_session, conf_lb_id=CONF.lb_id,
                                               conf_project_id=CONF.project_id)
    tenant_bindings_to_delete = []
    for lb_id in lb_id_list:
        try:
            lb_id = lb_id[0]
            # TODO: Preform a lookup of the associated device and cache it's name 
            # and associated tenant_id
            LOG.info('Locking load balancer: %s', lb_id)
            db_utils.lock_loadbalancer(n_session, lb_id)

            n_lb = db_utils.get_loadbalancer_entry(n_session, lb_id)
            provider = n_lb[0]
            tenant_id = n_lb[1]
            if provider != 'a10networks':
                LOG.info('Skipping loadbalancer with provider %s. Not an A10 Networks LB', provider)
                continue

            if a10_config.get('use_database'):
                device_name = aten2oct.get_device_name_by_tenant(a10_nlbaas_session, tenant_id)
            else:
                devices = a10_config.get('devices')
                device_name = acos_client.Hash(list(devices)).get_server(tenant_id)

            LOG.info('Migrating Thunder device: %s', device_name)
            device_info = a10_config.get_device(device_name)
            try:
                aten2oct.migrate_thunder(a10_oct_session, lb_id, tenant_id, device_info)
            except aten2oct.UnsupportedAXAPIVersionException as e:
                LOG.warning('Skipping loadbalancer %s for device %s with AXAPI version %s. '
                            'Only AXAPI version 3.0 is supported.',
                            lb_id, device_name, e.axapi_version)

            LOG.info('Migrating VIP port for load balancer: %s', lb_id)
            lb2oct.migrate_vip_ports(n_session, CONF.migration.octavia_account_id, lb_id, n_lb)

            LOG.info('Migrating load balancer: %s', lb_id)
            lb2oct.migrate_lb(o_session, lb_id, n_lb, fl_id)

            LOG.info('Migrating VIP for load balancer: %s', lb_id)
            lb2oct.migrate_vip(n_session, o_session, lb_id, n_lb)


            # Start pool migration
            pools = db_utils.get_pool_entries_by_lb(n_session, lb_id)
            for pool in pools:
                pool_id = pool[0]
                pool_state = pool[7]
                LOG.debug('Migrating pool: %s', pool_id)
                if pool_state == 'DELETED':
                    continue
                elif pool_state != 'ACTIVE':
                    raise Exception(_('Pool is invalid state of %s.'), pool_state)
                lb2oct.migrate_pools(o_session, lb_id, n_lb, pool)

                hm_id = pool[5]
                if hm_id is not None:
                    LOG.debug('Migrating health manager: %s', hm_id)
                    hm = db_utils.get_healthmonitor(n_session, hm_id)
                    lb2oct.migrate_health_monitor(o_session, tenant_id, pool_id, hm_id, hm)

                # Handle the session persistence records
                sp = db_utils.get_sess_pers_by_pool(n_session, pool_id)
                if sp:
                    LOG.debug('Migrating session persistence for pool: %s', pool_id)
                    lb2oct.migrate_session_persistence(o_session, pool_id, sp)

                # Handle the pool members
                members = db_utils.get_members_by_pool(n_session, pool_id)
                for member in members:
                    member_id = member[0]
                    member_state = member[6]
                    LOG.debug('Migrating member: %s', member_id)
                    if member_state == 'DELETED':
                        continue
                    elif member_state != 'ACTIVE':
                        raise Exception(_('Member %s for pool %s is invalid state of %s.'),
                                        member_id,
                                        pool_id,
                                        member_state)
                    lb2oct.migrate_member(o_session, tenant_id, pool_id, member)

            # Start listener migration. Must come after pool due to l7policy fk
            listeners, lb_stats = db_utils.get_listeners_and_stats_by_lb(n_session, lb_id)
            for listener in listeners:
                listener_id = listener[0]
                listener_state = listener[8]
                LOG.debug('Migrating listener: %s', listener_id)
                if listener_state == 'DELETED':
                    continue
                elif listener_state != 'ACTIVE':
                    raise Exception(_('Listener is invalid state of %s.'),
                                     listener_state)
                lb2oct.migrate_listener(n_session, o_session, lb_id, n_lb, listener, lb_stats)

                # Handle SNI certs
                SNIs = db_utils.get_SNIs_by_listener(n_session, listener_id)
                for SNI in SNIs:
                    sni_id = SNI[0]
                    LOG.debug('Migrating SNI: %s', sni_id)
                    lb2oct.migrate_SNI(o_session, listener_id, SNI)

                # Handle L7 policy records
                l7policies = db_utils.get_l7policies_by_listener(n_session, listener_id)
                for l7policy in l7policies:
                    l7policy_id = l7policy[0]
                    l7policy_state = l7policy[8]
                    LOG.debug('Migrating L7 policy: %s', l7policy_id)
                    if l7policy_state == 'DELETED':
                        continue
                    elif l7policy_state != 'ACTIVE':
                        raise Exception(_('L7 policy is invalid state of %s.'),
                                        l7policy_state)                    
                    lb2oct.migrate_l7policy(o_session, tenant_id, listener_id, l7policy)
                    
                     # Handle L7 rule records
                    l7rules = db_utils.get_l7rules_by_l7policy(n_session, l7policy_id)
                    for l7rule in l7rules:
                        l7rule_id = l7rule[0]
                        l7rule_state = l7rule[6]
                        LOG.debug('Migrating L7 rule: %s', l7rule_id)
                        if l7rule_state == 'DELETED':
                            continue
                        elif l7rule_state != 'ACTIVE':
                            raise Exception(_('L7 rule is invalid state of %s.'),
                                            l7rule_state)
                        lb2oct.migrate_l7rule(o_session, tenant_id, l7policy, l7rule)

            # Delete the old neutron-lbaas records
            if (CONF.migration.delete_after_migration and not
                    CONF.migration.trial_run):
                LOG.info('Performing cascading delete on loadbalancer %s.', lb_id)
                db_utils.cascade_delete_neutron_lb(n_session, lb_id)
                LOG.info('Successful cascading delete of loadbalancer %s.', lb_id)
                tenant_bindings_to_delete.append(tenant_id)
            
            # Rollback everything if we are in a trial run otherwise commit
            if CONF.migration.trial_run:
                o_session.rollback()
                n_session.rollback()
                LOG.info('Simulated migration of load balancer %s successful.',
                         lb_id)
            else:
                o_session.commit()
                n_session.commit()
                LOG.info('Successful migration of load balancer %s.', lb_id)
        except Exception as e:
            n_session.rollback()
            o_session.rollback()
            LOG.exception("Skipping load balancer %s due to: %s.", lb_id, str(e))
            failure_count += 1
        finally:
            # Attempt to unlock the loadbalancer even if an error occured or it was deleted.
            # This ensures we don't get stuck in pending states
            LOG.info('Unlocking load balancer: %s', lb_id)
            db_utils.unlock_loadbalancer(n_session, lb_id)
            n_session.commit()

    try:
        # We can't be sure when no more loadbalancers with a given tenant exist
        # in the DB. So we have to delete them here.
        if a10_config.get('use_database'):
            for tenant_binding in tenant_bindings_to_delete:
                LOG.info('Deleting A10 tenant biding for tenant: %s', tenant_binding)
                aten2oct.delete_binding_by_tenant(n_session, tenant_binding)
            if CONF.migration.trial_run:
                n_session.rollback()
                LOG.info('Simulated deletion of A10 tenant bindings successful.')
            elif len(tenant_bindings_to_delete) > 0:
                n_session.commit()
                LOG.info('Deletion of A10 tenant bindings successful')
    except Exception as e:
        n_session.rollback()
        LOG.exception("Skipping A10 tenant binding deletion due to: %s.", str(e))
        failure_count += 1

    if failure_count:
        LOG.warning("%d failures were detected", failure_count)
        sys.exit(1)

if __name__ == "__main__":
    main()



# Copyright 2020 A10 Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import sys

from oslo_db.sqlalchemy import enginefacade
import oslo_i18n as i18n
from oslo_log import log as logging
from oslo_utils import uuidutils


def _create_flavor_expr(expr_data):
    name_exprs = {'name-expressions': []}
    for k, v in expr_data.items():
        v['description'] = k
        name_exprs['name-expressions'].append(v)
    return name_exprs


def create_flavor_data(a10_config):
    flavor_data = {}

    vs_expr = a10_config.get_virtual_server_expressions()
    if vs_expr:
        flavor_data['virtual-server'] =  _create_flavor_expr(vs_expr)

    vport_expr = a10_config.get_vport_expressions()
    if vport_expr:
        flavor_data['virtual-port'] = _create_flavor_expr(vport_expr)

    hm_expr = a10_config.get_monitor_expressions()
    if hm_expr:
        flavor_data['health-monitor'] = _create_flavor_expr(hm_expr)


    sg_expr = a10_config.get_service_group_expressions()
    if sg_expr:
        flavor_data['service-group'] = _create_flavor_expr(sg_expr)

    mem_expr = a10_config.get_member_expressions()
    if mem_expr:
        flavor_data['server'] = _create_flavor_expr(mem_expr)

    return flavor_data


def create_flavorprofile(o_session, flavor_data):
    flavorprofile_id = uuidutils.generate_uuid()
    flavorprofile_name = flavorprofile_id[:10]
    provider_name = 'a10'

    flavor_data = str(flavor_data).replace('\'', '\"')
    result = o_session.execute(
        "INSERT INTO flavor_profile (id, name, provider_name, "
        "flavor_data) VALUES (:id, :name, :provider_name, :flavor_data);",
        {'id': flavorprofile_id, 'name': flavorprofile_name,
         'provider_name': provider_name, 'flavor_data': flavor_data})
    if result.rowcount != 1:
        raise Exception(_('Unable to create flavorprofile in the '
                        'Octavia database.'))
    return flavorprofile_id


def create_flavor(o_session, flavorprofile_id):
    flavor_id = uuidutils.generate_uuid()
    flavor_name = flavor_id[:10]

    result = o_session.execute(
        "INSERT INTO flavor (id, name, enabled, flavor_profile_id) "
        "VALUES (:id, :name, :enabled, :flavor_profile_id);",
        {'id': flavor_id, 'name': flavor_name, 'enabled': 1,
         'flavor_profile_id': flavorprofile_id})
    if result.rowcount != 1:
        raise Exception(_('Unable to create flavor in the '
                        'Octavia database.'))
    return flavor_id



# Copyright 2020 A10 Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import functools32

import oslo_i18n as i18n
from oslo_utils import uuidutils

_translators = i18n.TranslatorFactory(domain='a10_nlbaas2oct')

# The primary translation function using the well-known name "_"
_ = _translators.primary


class IncorrectPartitionTypeException(Exception):

    def __init__(self, v_method):
        self.message = ("v_method of type {} was specified, "
                       "but only \"LSI\" or \"ADP\" are supported")
        self.message = _(self.message.format(v_method)) # Apply translator
        super(IncorrectPartitionTypeException, self).__init__(self.message)


class UnsupportedAXAPIVersionException(Exception):

    def __init__(self, axapi_version):
        self.axapi_version = axapi_version
        super(IncorrectPartitionTypeException, self).__init__(self.message)


def get_device_name_by_tenant(a10_nlbaas_session, tenant_id):

    # The db session will change each time so we need this internal function instead
    @functools32.lru_cache(maxsize=None)
    def preform_db_select(db_tenant_id):
        # To avoid side effects, pass the tenant_id to the internal function
        # instead of using external definition
        device_name = a10_nlbaas_session.execute(
            "SELECT device_name FROM neutron.a10_tenant_bindings WHERE "
            "tenant_id = :tenant_id ;", {"tenant_id": db_tenant_id}).fetchone()
        return device_name[0]
    
    return preform_db_select(tenant_id)


def delete_binding_by_tenant(a10_nlbaas_session, tenant_id):
    # Delete the bindings
    a10_nlbaas_session.execute(
        "DELETE FROM neutron.a10_tenant_bindings WHERE tenant_id = :tenant_id;",
        {'tenant_id': tenant_id})


def migrate_thunder(a10_oct_session, loadbalancer_id, tenant_id, device_info):
    # Create thunder entry

    vthunder_id = uuidutils.generate_uuid()

    if device_info['v_method'] == "LSI":
        hierarchical_multitenancy = "disable"
        partition_name = device_info['shared_partition']
    elif device_info['v_method'] == "ADP":
        hierarchical_multitenancy = "enable"
        partition_name = tenant_id[0:13]
    else:
        raise IncorrectPartitionTypeException(device_info['v_method'])

    axapi_version = 30
    if device_info['api_version'] not in ['3.0', 30]:
        raise UnsupportedAXAPIVersionException(device_info['api_version'])

    result = a10_oct_session.execute(
        "INSERT INTO vthunders (vthunder_id, device_name, ip_address, username, "
        "password, axapi_version, undercloud, loadbalancer_id, project_id, "
        "topology, role, last_udp_update, status, created_at, updated_at, "
        "partition_name, hierarchical_multitenancy) "
        "VALUES (:vthunder_id, :device_name, :ip_address, :username, :password, "
        ":axapi_version, :undercloud, :loadbalancer_id, :project_id, :topology, "
        ":role, :last_udp_update, :status, :created_at, :updated_at, :partition_name, "
        ":hierarchical_multitenancy);",
        {'vthunder_id': vthunder_id,
         'device_name': device_info['name'],
         'ip_address': device_info['host'],
         'username': device_info['username'],
         'password': device_info['password'],
         'axapi_version': axapi_version,
         'undercloud': 1,
         'loadbalancer_id': loadbalancer_id,
         'project_id': tenant_id,
         'topology': "STANDALONE",
         'role': "MASTER",
         'status': "ACTIVE",
         'last_udp_update': datetime.datetime.utcnow(),
         'created_at': datetime.datetime.utcnow(),
         'updated_at': datetime.datetime.utcnow(),
         'partition_name': partition_name,
         'hierarchical_multitenancy': hierarchical_multitenancy}
        )
    if result.rowcount != 1:
        raise Exception(_('Unable to create Thunder in the A10 Octavia database.'))
